#include <inttypes.h>
#include <stdio.h>
#include <string.h>

#include "net/netdev2.h"

#include "net/gnrc/netif.h"
#include "net/gnrc/netreg.h"
#include "net/gnrc/netapi.h"
#include "net/gnrc/pktbuf.h"
#include "net/gnrc/netif/hdr.h"

#include "msg.h"
#include "thread.h"

static char rcv_stack[THREAD_STACKSIZE_MAIN];
static char serial_input_handler_stack[THREAD_STACKSIZE_MAIN];

static kernel_pid_t rcv_pid;
static kernel_pid_t serial_input_handler_pid;
static kernel_pid_t dev;

#define RCV_QUEUE_SIZE (8)
#define EVAL_PRR_NETTYPE GNRC_NETTYPE_UNDEF

static msg_t rcv_queue[RCV_QUEUE_SIZE];
static char buf[255];

static void _dump(gnrc_pktsnip_t *pkt) {
	//gnrc_pktsnip_t *pkt = msg.content.ptr;

	uint8_t* data = (uint8_t*) pkt->data;
	char* str = (char*) &data[1];

	printf("Packet(%p) #%d: %s\n", pkt, data[0], str);

	/*
	printf("Incoming packet(%p) (next=%p) of type %d, used by %d users and length %d: ", pkt, pkt->next, pkt->type, pkt->users, pkt->size);

	int i = 0;
	uint8_t* data = (uint8_t*) pkt->data;
	for(; i < pkt->size; i++) {
		printf(" %02X(%c)", data[i], data[i]);
	}
	printf("\n");
	*/

	//printf("PKTDUMP: data received:%p\n", pkt);

	// (void)pkt;
	gnrc_pktbuf_release(pkt);
}

void *rcv(void *arg)
{
	msg_t msg, reply;
	//(void*)arg;
	msg_init_queue(rcv_queue, RCV_QUEUE_SIZE);
  reply.type = GNRC_NETAPI_MSG_TYPE_ACK;
  reply.content.value = (uint32_t)-ENOTSUP;

	printf("rcv_pid: %d\n", thread_getpid());

  gnrc_netreg_entry_t me_reg = { .demux_ctx = GNRC_NETREG_DEMUX_CTX_ALL, .pid = thread_getpid() };
	gnrc_netreg_register(GNRC_NETTYPE_UNDEF, &me_reg);

	printf("rcv thread started\n");

  while (1) {
      msg_receive(&msg);
      switch (msg.type) {
          case GNRC_NETAPI_MSG_TYPE_RCV:
						_dump(msg.content.ptr);
          	break;
							/*
          case GNRC_NETAPI_MSG_TYPE_SND:
              pkt = msg.content.ptr;
							puts("Outgoing packet\n");
							(void)pkt;
							gnrc_pktbuf_release(pkt);
              break;
							*/
           case GNRC_NETAPI_MSG_TYPE_SET:
           case GNRC_NETAPI_MSG_TYPE_GET:
              msg_reply(&msg, &reply);
              break;
          default:
						puts("Received something unexpected...");
              break;
      }
  }

  return NULL;
}

static void* send_message(uint8_t seq_nr)
{
	gnrc_pktsnip_t *hdr;
	gnrc_pktsnip_t *pkt;
	gnrc_netif_hdr_t *nethdr;
	/// Not used in our case
	uint8_t dst[2] = {0xff, 0xff};
	uint8_t flags = 0x00;

	// For now send broadcast
	flags |= GNRC_NETIF_HDR_FLAGS_BROADCAST;

	char data[7] = {seq_nr, 'h', 'a', 'l', 'l', 'o', '?'};
	pkt = gnrc_pktbuf_add(NULL, data, sizeof(data), GNRC_NETTYPE_UNDEF); // EVAL_PRR_NETTYPE
	/// NULL/0: no src address
	hdr = gnrc_netif_hdr_build(NULL, 0, dst, 0);
	if (hdr == NULL) {
			puts("error: packet buffer full");
			gnrc_pktbuf_release(pkt);
			return 0;
	}
	LL_PREPEND(pkt, hdr);
	nethdr = (gnrc_netif_hdr_t*) hdr->data;
	nethdr->flags = flags;
	if (pkt == NULL) {
	     puts("Error: unable to copy data to packet buffer\n");
	} else if (gnrc_netapi_send(dev, pkt) < 1) { // GNRC_NETTYPE_UNDEF, GNRC_NETREG_DEMUX_CTX_ALL
		// gnrc_netapi_dispatch_send(GNRC_NETTYPE_UNDEF, GNRC_NETREG_DEMUX_CTX_ALL, pkt) < 1
	     puts("Error: no thread is interested\n");
	     gnrc_pktbuf_release(pkt);
	}

	return NULL;
}

static void* serial_input_handler(void *arg) {
	char *ptr = buf;
	while(1) {
		int c = getchar();
		if(c < 0) {
			continue;
		}
		/// Check if received a newline character
		if(c == '\n') {
			// if 2nd last character is a carriage return
			if(*(ptr-1) == '\r') {
				ptr--;
			}
			*ptr = '\0';
			if(ptr != buf) {
				//uint8_t len = (ptr - buf);
				if(strcmp(buf, "s") == 0) {
					for(uint8_t i=0;i<10;i++) {
						send_message(i);
						netopt_state_t state;
						/// Wait until packet has been send
						while(gnrc_netapi_get(dev, NETOPT_STATE, 0, &state, sizeof(state)) == NETOPT_STATE_RX);
					}
				}
			}
			/// Reset ptr
			ptr = buf;
			continue;
		}

		*ptr++ = c;
	}

	return NULL;
}


int main(void)
{
	kernel_pid_t ifs[GNRC_NETIF_NUMOF];

	size_t numof = gnrc_netif_get(ifs);

	if(numof == 0) {
		puts("No network interfaces found");
		return 0;
	}

	printf("found %d interface(s). First: %d\n", numof, ifs[0]);

	dev = ifs[0];

	/*
  if (numof > 0) {
      gnrc_ipv6_netif_t *entry = gnrc_ipv6_netif_get(ifs[0]);
      for (int i = 0; i < GNRC_IPV6_NETIF_ADDR_NUMOF; i++) {
          if ((ipv6_addr_is_link_local(&entry->addrs[i].addr)) && !(entry->addrs[i].flags & GNRC_IPV6_NETIF_ADDR_FLAGS_NON_UNICAST)) {
              char ipv6_addr[IPV6_ADDR_MAX_STR_LEN];
              ipv6_addr_to_str(ipv6_addr, &entry->addrs[i].addr, IPV6_ADDR_MAX_STR_LEN);
              printf("My address is %s\n", ipv6_addr);
          }
      }
  } else {
		puts("No addresses found :(");
	}
	*/

	uint16_t channel = 25;
	gnrc_netapi_set(dev, NETOPT_CHANNEL, 0, &channel, sizeof(channel));

	rcv_pid = thread_create(rcv_stack, sizeof(rcv_stack), THREAD_PRIORITY_MAIN - 1, 0, rcv, NULL, "rcv");
	serial_input_handler_pid = thread_create(serial_input_handler_stack, sizeof(serial_input_handler_stack), THREAD_PRIORITY_MIN, 0, serial_input_handler, NULL, "serial_input_handler");

	printf("rcv_pid(main): %d\n", rcv_pid);

	return 0;
}
